import click
import keyring

# importing the backends explicitely is required for PyInstaller to work
import keyring.backends.kwallet
import keyring.backends.Windows
import keyring.backends.OS_X
import keyring.backends.SecretService
import keyring.backends.chainer

from . import base
from .. import api


class lazy_credential:
    """
    A proxy object to request access to the proxy object at the later possible point,
    cf #4
    """

    def __init__(self, *args):
        self.args = args
        self._cached_value = None

    @property
    def value(self):
        if self._cached_value:
            return self._cached_value
        try:
            v = keyring.get_password(*self.args)
        except ValueError as e:
            raise click.ClickException(
                "Error while retrieving password from keyring: {}. Your password may be incorrect.".format(
                    e.args[0]
                )
            )
        except Exception as e:
            raise click.ClickException(
                "Error while retrieving password from keyring: {}".format(e.args[0])
            )
        self._cached_value = v
        return v

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        return self.value == other

    def __repr__(self):
        return str(self.value)

    def __bool__(self):
        return bool(self.value)

    def __len__(self):
        return len(str(self))

def init_keyring():
    # small hack to fix some weird issues with pyinstaller and keyring
    # there seems to be a cache issue somewhere
    try:
        del keyring.backend.get_all_keyring.__wrapped__.always_returns
    except AttributeError:
        pass
    keyring.core.init_backend()
    # /end of hack


@base.cli.command()
@click.option("-u", "--username", envvar="FUNKWHALE_USERNAME", prompt=True)
@click.option(
    "-p", "--password", envvar="FUNKWHALE_PASSWORD", prompt=True, hide_input=True
)
@click.pass_context
@base.async_command
async def login(ctx, username, password):
    async with api.get_session() as session:
        token = await api.get_jwt_token(
            session, ctx.obj["SERVER_URL"], username=username, password=password
        )

    try:
        keyring.set_password(ctx.obj["SERVER_URL"], "_", token)
    except ValueError as e:
        raise click.ClickException(
            "Error while retrieving password from keyring: {}. Your password may be incorrect.".format(
                e.args[0]
            )
        )
    except Exception as e:
        raise click.ClickException(
            "Error while retrieving password from keyring: {}".format(e.args[0])
        )
    click.echo("Login successfull!")


@base.cli.command()
@click.pass_context
@base.async_command
async def logout(ctx):
    keyring.delete_password(ctx.obj["SERVER_URL"], "_")
    click.echo("Logout successfull!")
