import asyncio
import aiohttp
import click
import click_log
import dotenv
import functools
import ssl
import sys
import urllib.parse

import math
import json

from funkwhale_cli import api
from funkwhale_cli import config
from funkwhale_cli import exceptions
from funkwhale_cli import logs
from funkwhale_cli import output
from funkwhale_cli import utils

click_log.basic_config(logs.logger)

NOOP = object()

SSL_PROTOCOLS = (asyncio.sslproto.SSLProtocol,)
try:
    import uvloop.loop
except ImportError:
    pass
else:
    SSL_PROTOCOLS = (*SSL_PROTOCOLS, uvloop.loop.SSLProtocol)


def ignore_aiohttp_ssl_eror(loop):
    """Ignore aiohttp #3535 / cpython #13548 issue with SSL data after close

    There is an issue in Python 3.7 up to 3.7.3 that over-reports a
    ssl.SSLError fatal error (ssl.SSLError: [SSL: KRB5_S_INIT] application data
    after close notify (_ssl.c:2609)) after we are already done with the
    connection. See GitHub issues aio-libs/aiohttp#3535 and
    python/cpython#13548.

    Given a loop, this sets up an exception handler that ignores this specific
    exception, but passes everything else on to the previous exception handler
    this one replaces.

    Checks for fixed Python versions, disabling itself when running on 3.7.4+
    or 3.8.

    """
    if sys.version_info >= (3, 7, 4):
        return

    orig_handler = loop.get_exception_handler()

    def ignore_ssl_error(loop, context):
        if context.get("message") in {
            "SSL error in data received",
            "Fatal error on transport",
        }:
            # validate we have the right exception, transport and protocol
            exception = context.get("exception")
            protocol = context.get("protocol")
            if (
                isinstance(exception, ssl.SSLError)
                and exception.reason == "KRB5_S_INIT"
                and isinstance(protocol, SSL_PROTOCOLS)
            ):
                if loop.get_debug():
                    asyncio.log.logger.debug("Ignoring asyncio SSL KRB5_S_INIT error")
                return
        if orig_handler is not None:
            orig_handler(loop, context)
        else:
            loop.default_exception_handler(context)

    loop.set_exception_handler(ignore_ssl_error)


def noop_decorator(f):
    return f


def URL(v):
    if v is NOOP:
        raise click.ClickException(
            "You need to specify a server, either via the -H flag or using the FUNKWHALE_SERVER_URL environment variable"
        )
    v = str(v) if v else None
    parsed = urllib.parse.urlparse(v)
    if parsed.scheme not in ["http", "https"] or not parsed.netloc:
        raise ValueError("{} is not a valid url".format(v))
    if not v.endswith("/"):
        v = v + "/"
    return v


def env_file(v):
    if v is NOOP:
        v = None
    if v is not None:
        v = click.Path(exists=True)(v)
    env_files = [v or ".env", config.get_env_file()]
    for p in env_files:
        logs.logger.debug("Loading env file at {}".format(p))
        dotenv.load_dotenv(p, override=False)

    return v


def async_command(f):
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        ignore_aiohttp_ssl_eror(loop)
        _async_reraise = kwargs.pop("_async_reraise", False)
        try:
            return loop.run_until_complete(f(*args, **kwargs))
        except (aiohttp.client_exceptions.ClientError) as e:
            if _async_reraise:
                raise
            message = str(e)
            if hasattr(e, "status") and e.status == 401:
                message = "Remote answered with {}, ensure your are logged in first".format(
                    e.status
                )
            raise click.ClickException(message)
        except (exceptions.FunkwhaleError) as e:
            if _async_reraise:
                raise
            message = str(e)
            raise click.ClickException(message)
        else:
            raise

    return functools.update_wrapper(wrapper, f)


async def check_status(response):
    text = await response.text()
    try:
        response.raise_for_status()
    except aiohttp.client_exceptions.ClientError as e:
        raise click.ClickException(str(e) + ": {}".format(text))


SERVER_DECORATOR = click.option(
    "-H",
    "--url",
    envvar="FUNKWHALE_SERVER_URL",
    type=URL,
    default=NOOP,
    help="The URL of the Funkwhale server to query",
)
TOKEN_DECORATOR = click.option(
    "-t",
    "--token",
    envvar="FUNKWHALE_TOKEN",
    help="A Bearer token to use for authentication",
)
RAW_DECORATOR = click.option(
    "--raw", is_flag=True, help="Directly output JSON returned by the happy"
)


def set_server(ctx, url, token, use_auth=True):
    from . import auth

    ctx.ensure_object(dict)
    ctx.obj["SERVER_URL"] = url
    parsed = urllib.parse.urlparse(url)
    ctx.obj["SERVER_NETLOC"] = parsed.netloc
    ctx.obj["SERVER_PROTOCOL"] = parsed.scheme
    token = (token or auth.lazy_credential(url, "_")) if use_auth else None
    ctx.obj["remote"] = api.get_api(
        domain=ctx.obj["SERVER_NETLOC"],
        protocol=ctx.obj["SERVER_PROTOCOL"],
        token=token,
    )


@click.group()
@click.option(
    "-e",
    "--env-file",
    envvar="ENV_FILE",
    type=env_file,
    default=NOOP,
    help="Path to an env file to use. A .env file will be used automatically if any",
)
@click.option(
    "-q",
    "--quiet",
    envvar="FUNKWHALE_QUIET",
    is_flag=True,
    default=False,
    help="Disable logging",
)
@click.option(
    "--no-login",
    envvar="FUNKWHALE_NO_LOGIN",
    is_flag=True,
    default=False,
    help="Disable authentication/keyring",
)
@SERVER_DECORATOR
@TOKEN_DECORATOR
@click_log.simple_verbosity_option(logs.logger, expose_value=True)
@click.pass_context
def cli(ctx, env_file, url, verbosity, token, quiet, no_login):
    from . import auth

    auth.init_keyring()
    ctx.ensure_object(dict)
    logs.logger.disabled = quiet
    set_server(ctx, url, token, use_auth=not no_login)


def get_pagination_data(payload):
    data = {"next_page": None, "page_size": None}
    if payload.get("next"):
        next_page = utils.get_url_param(payload["next"], "page")
        data["next_page"] = int(next_page)
        data["total_pages"] = math.ceil(payload["count"] / len(payload["results"]))
        data["current_page"] = int(next_page) - 1
        data["page_size"] = len(payload["results"])

    if payload.get("previous"):
        previous_page = utils.get_url_param(payload["previous"], "page") or 0
        data.setdefault("current_page", int(previous_page) + 1)
        data.setdefault("total_pages", data["current_page"])
        if (
            not data["page_size"]
            and payload["count"] - len(payload["results"]) > 0
            and data["total_pages"] > 1
        ):
            data["page_size"] = int(payload["count"] - len(payload["results"])) / (
                data["total_pages"] - 1
            )
    data.setdefault("current_page", 1)
    data.setdefault("total_pages", 1)
    return data


def get_ls_command(
    group,
    endpoint,
    output_conf,
    pagination=True,
    filter=True,
    ordering=True,
    scope=False,
    with_id=False,
    owned_conf=None,
    name="ls",
    doc="",
    id_metavar="ID",
):

    available_fields = sorted(
        set(output_conf["labels"]) | set(output.FIELDS["*"].keys())
    )
    id_decorator = (
        click.argument("id", metavar=id_metavar) if with_id else noop_decorator
    )
    page_decorator = (
        click.option("--page", "-p", type=click.INT, default=1)
        if pagination
        else noop_decorator
    )
    page_size_decorator = (
        click.option("--page-size", "-s", type=click.INT, default=None)
        if pagination
        else noop_decorator
    )
    limit_decorator = (
        click.option("--limit", "-l", type=click.INT, default=1)
        if pagination
        else noop_decorator
    )
    ordering_decorator = (
        click.option("--ordering", "-o", default=None) if ordering else noop_decorator
    )

    scope_decorator = click.option("--scope", default=None) if scope else noop_decorator
    filter_decorator = (
        click.option("--filter", "-f", multiple=True) if filter else noop_decorator
    )
    owned_decorator = (
        click.option("--owned", is_flag=True, default=False)
        if owned_conf
        else noop_decorator
    )

    @id_decorator
    @click.argument("query", nargs=-1)
    @RAW_DECORATOR
    @click.option(
        "--format", "-t", type=click.Choice(output.TABLE_FORMATS), default="simple"
    )
    @click.option("--no-headers", "-h", is_flag=True, default=False)
    @click.option("--ids", "-i", is_flag=True)
    @page_decorator
    @page_size_decorator
    @ordering_decorator
    @scope_decorator
    @filter_decorator
    @limit_decorator
    @owned_decorator
    @click.option(
        "--column",
        "-c",
        multiple=True,
        help="Which column to display. Available: {}. \nDefault: {}".format(
            ", ".join(available_fields), ", ".join(output_conf["labels"])
        ),
    )
    @click.pass_context
    @async_command
    async def ls(ctx, raw, column, format, no_headers, ids, **kwargs):
        id = kwargs.get("id")
        limit = kwargs.get("limit")
        page = kwargs.get("page")
        page_size = kwargs.get("page_size")
        ordering = kwargs.get("ordering")
        scope = kwargs.get("scope")
        filter = kwargs.get("filter")
        query = kwargs.get("query")
        owned = kwargs.get("owned")
        if ids:
            no_headers = True
            column = [output_conf.get("id_field", "UUID")]
            format = "plain"
        base_url = endpoint
        if with_id:
            base_url = base_url.format(id)
        next_page_url = None
        page_count = 0
        while True:
            if limit and page_count >= limit:
                break
            async with ctx.obj["remote"]:
                if not pagination or page_count == 0:
                    url = base_url
                    params = {}
                    if page:
                        params["page"] = page
                    if page_size:
                        params["page_size"] = page_size
                    if ordering:
                        params["ordering"] = ordering
                    if scope:
                        params["scope"] = scope
                    if query:
                        params["q"] = " ".join(query)
                    if filter:
                        for f in filter:
                            query = urllib.parse.parse_qs(f)
                            for k, v in query.items():
                                params[k] = v[0]
                    if owned_conf and owned:
                        user_info = await get_user_info(ctx)
                        params[owned_conf["param"]] = utils.recursive_getattr(
                            user_info, owned_conf["field"]
                        )

                else:
                    params = {}
                    url = next_page_url
                if not url:
                    break
                result = await ctx.obj["remote"].request("get", url, params=params)
                result.raise_for_status()
                payload = await result.json()
                next_page_url = payload.get("next")
                page_count += 1
            if raw:
                click.echo(json.dumps(payload, sort_keys=True, indent=4))
            else:
                click.echo(
                    output.table(
                        payload["results"],
                        column or output_conf["labels"],
                        type=output_conf["type"],
                        format=format,
                        headers=not no_headers,
                    )
                )
                if not pagination:
                    break
                pagination_data = get_pagination_data(payload)
                if pagination_data["page_size"]:
                    start = (
                        int(
                            (pagination_data["current_page"] - 1)
                            * pagination_data["page_size"]
                        )
                        + 1
                    )
                else:
                    start = 1
                end = min(start + len(payload["results"]) - 1, payload["count"])
                logs.logger.info(
                    "\nObjects {start}-{end} on {total} (page {current_page}/{total_pages})".format(
                        start=start,
                        end=end,
                        total=payload["count"],
                        current_page=pagination_data["current_page"],
                        total_pages=pagination_data["total_pages"] or 1,
                    )
                )

    ls.__doc__ = doc
    return group.command(name)(ls)


def get_show_command(
    group, url_template, output_conf, name="show", force_id=None, doc=""
):

    available_fields = sorted(
        set(output_conf["labels"]) | set(output.FIELDS["*"].keys())
    )
    if force_id:

        def id_decorator(f):
            @functools.wraps(f)
            def inner(raw, column, format):
                return f(raw, force_id, column, format)

            return inner

    else:
        id_decorator = click.argument("id")

    @id_decorator
    @RAW_DECORATOR
    @click.option(
        "--format", "-t", type=click.Choice(output.TABLE_FORMATS), default="simple"
    )
    @click.option(
        "--column",
        "-c",
        multiple=True,
        help="Which column to display. Available: {}. \nDefault: {}".format(
            ", ".join(available_fields), ", ".join(output_conf["labels"])
        ),
    )
    @click.pass_context
    @async_command
    async def show(ctx, raw, id, column, format):

        async with ctx.obj["remote"]:
            async with ctx.obj["remote"].request(
                "get", url_template.format(id)
            ) as result:
                result.raise_for_status()
                payload = await result.json()
        if raw:
            click.echo(json.dumps(payload, sort_keys=True, indent=4))
        else:
            click.echo(
                output.obj_table(
                    payload,
                    column or output_conf["labels"],
                    type=output_conf["type"],
                    format=format,
                )
            )

    show.__doc__ = ""
    return group.command(name)(show)


def get_delete_command(
    group,
    url_template,
    confirm="Do you want to delete {} objects? This action is irreversible.",
    doc="Delect the given objects",
    name="rm",
    id_metavar="ID",
):
    @click.argument("id", nargs=-1, metavar=id_metavar)
    @RAW_DECORATOR
    @click.option("--no-input", is_flag=True)
    @click.pass_context
    @async_command
    async def delete(ctx, raw, id, no_input):
        async with ctx.obj["remote"]:
            if not no_input and not click.confirm(confirm.format(len(id)), abort=True):
                return
            for i in id:
                result = await ctx.obj["remote"].request(
                    "delete", url_template.format(i)
                )
                if result.status == 404:
                    logs.logger.warn("Couldn't delete {}: object not found".format(i))
                else:
                    result.raise_for_status()

        click.echo("{} Objects deleted!".format(len(id)))

    delete.__doc__ = doc
    return group.command(name)(delete)


async def get_user_info(ctx):
    async with ctx.obj["remote"].request("get", "api/v1/users/users/me/") as result:
        result.raise_for_status()
        return await result.json()
