import click
import json

from . import base
from .. import output


@base.cli.group()
@click.pass_context
def channels(ctx):
    """
    Manage channels
    """


channels_ls = base.get_ls_command(
    channels,
    "api/v1/channels/",
    scope=True,
    output_conf={
        "labels": ["UUID", "Name", "Category", "Username", "Tags"],
        "type": "CHANNEL",
    },
)
channels_rm = base.get_delete_command(channels, "api/v1/channels/{}/")


@channels.command("create")
@click.option("--name", prompt=True)
@click.option("--username", prompt=True)
@click.option("--cover")
@click.option("--description")
@click.option("--tags", default="")
@click.option("--itunes-category", default=None)
@click.option("--language", default=None)
@click.option(
    "--content-category",
    type=click.Choice(["music", "podcast", "other"]),
    default="podcast",
    prompt=True,
)
@click.option("--raw", is_flag=True)
@click.pass_context
@base.async_command
async def channels_create(
    ctx,
    raw,
    name,
    username,
    cover,
    description,
    tags,
    language,
    itunes_category,
    content_category,
):
    data = {
        "name": name,
        "username": username,
        "content_category": content_category,
    }
    if description:
        data["description"] = {
            "text": description,
            "content_type": "text/markdown",
        }
    else:
        data["description"] = None

    if cover:
        data["cover"] = cover

    missing_fields = []
    if not itunes_category:
        missing_fields.append(("--itunes-category", "itunes_category"))
    if not language:
        missing_fields.append(("--language", "language"))

    if content_category == "podcast" and missing_fields:
        click.secho(
            "You must provide adequate values for these fields: {}. Allowed values are listed below".format(
                ", ".join([f for f, _ in missing_fields])
            ),
            fg="red",
        )

        async with ctx.obj["remote"]:
            result = await ctx.obj["remote"].request(
                "get", "api/v1/channels/metadata-choices"
            )
            choices = await result.json()

        for flag, field in missing_fields:
            click.echo(
                "{}: {}".format(flag, ", ".join([c["value"] for c in choices[field]]))
            )

        raise click.ClickException("Missing params")
    data["tags"] = [t.strip() for t in tags.replace(" ", ",").split(",") if t]
    if content_category == "podcast":
        data["metadata"] = {
            "itunes_category": itunes_category,
            "language": language,
        }

    async with ctx.obj["remote"]:
        result = await ctx.obj["remote"].request("post", "api/v1/channels/", json=data)
        await base.check_status(result)
        payload = await result.json()
    if raw:
        click.echo(json.dumps(payload, sort_keys=True, indent=4))
    else:
        click.echo("Channel created:")
        click.echo(
            output.table(
                [payload], ["UUID", "Name", "Category", "Username"], type="CHANNEL"
            )
        )
