import click
import json

from . import base
from .. import output


@base.cli.group()
@click.pass_context
def playlists(ctx):
    """
    Manage playlists
    """


playlists_ls = base.get_ls_command(
    playlists,
    "api/v1/playlists/",
    doc="List available playlists",
    owned_conf={"param": "user", "field": "id"},
    output_conf={
        "labels": [
            "ID",
            "Name",
            "Visibility",
            "Tracks Count",
            "User",
            "Created",
            "Modified",
        ],
        "type": "PLAYLIST",
    },
)
playlists_rm = base.get_delete_command(
    playlists,
    "api/v1/playlists/{}/",
    id_metavar="PLAYLIST_ID",
    doc="Delete the given playlists",
)


@playlists.command("create")
@click.option("--name", prompt=True)
@click.option(
    "--visibility",
    type=click.Choice(["me", "instance", "everyone"]),
    default="me",
    prompt=True,
)
@click.option("--raw", is_flag=True)
@click.pass_context
@base.async_command
async def playlists_create(ctx, raw, name, visibility):
    """
    Create a new playlist
    """
    async with ctx.obj["remote"]:
        result = await ctx.obj["remote"].request(
            "post", "api/v1/playlists/", data={"name": name, "privacy_level": visibility}
        )
        result.raise_for_status()
        payload = await result.json()
    if raw:
        click.echo(json.dumps(payload, sort_keys=True, indent=4))
    else:
        click.echo("Playlist created:")
        click.echo(
            output.table(
                [payload], ["ID", "Name", "Visibility", "Tracks Count"], type="PLAYLIST"
            )
        )


@playlists.command("tracks-add")
@click.argument("id", metavar="PLAYLIST_ID")
@click.argument("track", nargs=-1, metavar="[TRACK_ID]…")
@click.option(
    "--no-duplicates",
    is_flag=True,
    default=False,
    help="Prevent insertion of tracks that already exist in the playlist. An error will be raised in this case.",
)
@click.pass_context
@base.async_command
async def playlists_tracks_add(ctx, id, track, no_duplicates):
    """
    Insert one or more tracks in a playlist
    """
    if not track:
        return click.echo("No track id provided")

    async with ctx.obj["remote"]:
        async with ctx.obj["remote"].request(
            "post",
            "api/v1/playlists/{}/add".format(id),
            data={"tracks": track, "allow_duplicates": not no_duplicates},
        ) as response:
            response.raise_for_status()


playlists_tracks = base.get_ls_command(
    playlists,
    "api/v1/playlists/{}/tracks/",
    name="tracks",
    with_id=True,
    pagination=False,
    ordering=False,
    filter=False,
    output_conf={
        "labels": ["Position", "ID", "Title", "Artist", "Album", "Created"],
        "id_field": "ID",
        "type": "PLAYLIST_TRACK",
    },
    id_metavar="PLAYLIST_ID",
    doc="List the tracks included in a playlist",
)
