import os
import pathlib

import aiohttp
import click
import tqdm

from . import base
from .. import logs
from .. import utils

TEMPLATE_ENV_VAR = "FUNKWHALE_DOWNLOAD_PATH_TEMPLATE"

TEMPLATE_DEFAULT = "{artist} - {album} - {title}.{extension}"


@base.cli.group()
@click.pass_context
def tracks(ctx):
    pass


tracks_ls = base.get_ls_command(
    tracks,
    "api/v1/tracks/",
    output_conf={
        "labels": ["ID", "Title", "Artist", "Album", "Disc", "Position"],
        "type": "TRACK",
        "id_field": "ID",
    },
)


async def get_track_download_url(id, remote, format=None):
    result = await remote.request("get", "api/v1/tracks/{}/".format(id))
    result.raise_for_status()
    payload = await result.json()

    try:
        download_url = payload["uploads"][0]["listen_url"]
    except IndexError:
        if remote.token:
            raise click.ClickException("This file is not available for download: " + id)
        else:
            raise click.ClickException(
                "This file is not available for download, try to login first: " + id
            )

    if download_url.startswith("/"):
        download_url = remote.base_url[:-1] + download_url
    if format:
        download_url = utils.add_url_params(download_url, {"to": format})
    else:
        format = payload["uploads"][0]["extension"]

    return download_url, format, payload


def extract_filename_params(track_data, format):
    filename_params = utils.flatten(track_data)
    filename_params["album"] = filename_params.get("album_title", None)
    filename_params["artist"] = filename_params.get("artist_name", None)
    filename_params["extension"] = format
    try:
        filename_params["year"] = filename_params["album_release_date"][:4]
    except (KeyError, TypeError):
        filename_params["year"] = None
    return {
        k: utils.sanitize_recursive(v) for k, v in filename_params.items()
    }


@tracks.command("download")
@click.argument("id", nargs=-1, required=True)
@click.option("--format", "-f")
@click.option("-d", "--directory", type=click.Path(exists=True))
@click.option("-o", "--overwrite", is_flag=True, default=False)
@click.option("-s", "--skip-existing", is_flag=True, default=False)
@click.option("-i", "--ignore-errors", multiple=True, type=int)
@click.option(
    "-t",
    "--template",
    default=TEMPLATE_DEFAULT,
    envvar=TEMPLATE_ENV_VAR,
)
@click.pass_context
@base.async_command
async def track_download(
        ctx, id, format, directory, template, overwrite, ignore_errors, skip_existing
):
    async with ctx.obj["remote"]:
        progressbar = tqdm.tqdm(id, unit="Files")
        for i in progressbar:
            try:
                download_url, extension, track_data = await get_track_download_url(
                    i, ctx.obj["remote"], format=format
                )
            except click.ClickException as e:
                logs.logger.error(e.message)
                continue
            logs.logger.info("Downloading from {}".format(download_url))

            filename_params = extract_filename_params(track_data, extension)

            if directory:
                filename = template.format(**filename_params)
                full_path = os.path.join(directory, filename)
                existing = os.path.exists(full_path)
                if skip_existing and existing:
                    logs.logger.info(
                        "'{}' already exists on disk, skipping download".format(
                            full_path
                        )
                    )
                    continue
                elif not overwrite and existing:
                    raise click.ClickException(
                        "'{}' already exists on disk. Relaunch this command with --overwrite if you want to replace it".format(
                            full_path
                        )
                    )

            async with ctx.obj["remote"].request(
                    "get", download_url, timeout=0
            ) as response:
                try:
                    response.raise_for_status()
                except aiohttp.ClientResponseError as e:
                    if response.status in ignore_errors:
                        logs.logger.warning(
                            "Remote answered with {} for url {}, skipping".format(
                                response.status, download_url
                            )
                        )
                        continue
                    else:
                        raise click.ClickException(
                            "Remote answered with {} for url {}, exiting".format(
                                response.status, download_url
                            )
                        )
                if directory:
                    final_directory = os.path.dirname(full_path)
                    pathlib.Path(final_directory).mkdir(parents=True, exist_ok=True)
                    logs.logger.info("Downloading to {}".format(full_path))
                    out = open(full_path, "wb")
                else:
                    out = click.get_binary_stream("stdout")
                while True:
                    chunk = await response.content.read(1024)
                    if not chunk:
                        break
                    out.write(chunk)
        logs.logger.info("Download complete")


async def get_track_data(id, remote, format=None):
    result = await remote.request("get", "api/v1/tracks/{}/".format(id))
    result.raise_for_status()
    payload = await result.json()

    if not format and payload["uploads"]:
        format = payload["uploads"][0]["extension"]

    return format, payload


@tracks.command("generate-playlist")
@click.argument("id", nargs=-1, required=True)
@click.option("--format", "-f")
@click.option("-d", "--directory", type=click.Path(exists=True))
@click.option(
    "-t",
    "--template",
    default=TEMPLATE_DEFAULT,
    envvar=TEMPLATE_ENV_VAR,
)
@click.option("-b", "--base-path", default="./")
@click.option("-n", "--name", default="playlist")
@click.pass_context
@base.async_command
async def track_generate_playlist(
        ctx, id, format, directory, template, base_path, name
):
    async with ctx.obj["remote"]:
        playlist = []
        for i in id:
            extension, track_data = await get_track_data(
                i, ctx.obj["remote"], format=format
            )

            filename_params = extract_filename_params(track_data, extension)
            filename = base_path + template.format(**filename_params)
            playlist.append(filename)
        if directory:
            filename = name + ".m3u8"
            full_path = os.path.join(directory, filename)
            logs.logger.info("Writing playlist to {}".format(full_path))
            with open(full_path, "w") as out:
                for track in playlist:
                    out.write(track + "\n")
        else:
            out = click.get_binary_stream("stdout")
            for track in playlist:
                out.write(bytes(track + "\n", "utf8"))
        logs.logger.info("Export of playlist complete")
