import datetime
import json
import os

import asyncio
import click
import tqdm

from . import base
from .. import logs


@base.cli.group()
@click.pass_context
def uploads(ctx):
    pass


uploads_ls = base.get_ls_command(  # noqa
    uploads,
    "api/v1/uploads/",
    output_conf={
        "labels": ["UUID", "Track", "Artist", "Import status", "Size", "Mimetype"],
        "type": "UPLOAD",
    },
)


def track_read(file_obj, name, progress):
    read = file_obj.read

    def patched_read(size):
        content = read(size)
        progress.update(len(content))
        progress.set_postfix(file=name[-30:], refresh=False)
        return content

    setattr(file_obj, "read", patched_read)


def get_valid_fields(metadata):
    data = {}
    for field in ["title", "position", "tags", "description"]:
        if field not in metadata:
            continue
        if field == "description":
            data[field] = metadata[field]["text"]
        else:
            data[field] = metadata[field]

    return data


async def upload(
    path,
    size,
    remote,
    ref,
    container_type,
    draft,
    album,
    license,
    library_or_channel_id,
    semaphore,
    global_progress,
):
    async with semaphore:
        filename = os.path.basename(path)
        data = {
            container_type: library_or_channel_id,
            "import_reference": ref,
            "source": "upload://{}".format(filename),
            "audio_file": open(path, "rb"),
        }
        if container_type == "channel":
            # needed to set proper metadata in the file later on before publication
            data["import_status"] = "draft"
            data["import_metadata"] = json.dumps(
                {"title": filename, "album": album, "license": license}
            )

        track_read(data["audio_file"], filename, global_progress)
        response = await remote.request("post", "api/v1/uploads/", data=data, timeout=0)
        upload = await base.check_status(response)
        upload = await response.json()
        if container_type == "channel":
            metadata_response = await remote.request(
                "get", "api/v1/uploads/{}/audio-file-metadata/".format(upload["uuid"])
            )
            metadata_response.raise_for_status()
            metadata = await metadata_response.json()
            new_data = {"import_metadata": upload["import_metadata"]}
            new_data["import_metadata"].update(get_valid_fields(metadata))
            if not draft:
                new_data["import_status"] = "pending"
            patch_response = await remote.request(
                "patch", "api/v1/uploads/{}/".format(upload["uuid"]), json=new_data
            )
            patch_response.raise_for_status()

        return response


@uploads.command("create")
@click.argument("library_or_channel_id")
@click.argument("paths", nargs=-1)
@click.option("-r", "--ref", default=None)
@click.option(
    "-a",
    "--album",
    default=None,
    help="Album to associate with the uploads. Only used when --channel is provided",
)
@click.option(
    "-l",
    "--license",
    default=None,
    help="License to associate with the uploads. Only used when --channel is provided",
)
@click.option(
    "-c",
    "--channel",
    is_flag=True,
    default=False,
    help="Provide this flag if you're uploading to a channel",
)
@click.option(
    "-d",
    "--draft",
    is_flag=True,
    default=False,
    help="Provide this flag if you want to upload in draft and publish later",
)
@click.option("-p", "--parallel", type=click.INT, default=1)
@click.pass_context
@base.async_command
async def uploads_create(
    ctx, library_or_channel_id, paths, ref, parallel, draft, channel, album, license
):
    logs.logger.info("Uploading {} files…".format(len(paths)))
    paths = sorted(set(paths))
    if not paths:
        return
    ref = ref or "funkwhale-cli-import-{}".format(datetime.datetime.now().isoformat())
    sizes = {path: os.path.getsize(path) for path in paths}

    async with ctx.obj["remote"]:
        if channel:
            logs.logger.info(
                "Checking channel {} existence…".format(library_or_channel_id)
            )
            channel_data = await ctx.obj["remote"].request(
                "get", "api/v1/channels/{}/".format(library_or_channel_id)
            )
            channel_data.raise_for_status()
        else:
            logs.logger.info(
                "Checking library {} existence…".format(library_or_channel_id)
            )
            library_data = await ctx.obj["remote"].request(
                "get", "api/v1/libraries/{}/".format(library_or_channel_id)
            )
            library_data.raise_for_status()
        sem = asyncio.Semaphore(parallel)
        progressbar = tqdm.tqdm(
            total=sum(sizes.values()), unit="B", unit_scale=True, unit_divisor=1024
        )

        tasks = [
            upload(
                path=path,
                ref=ref,
                size=sizes[path],
                draft=draft,
                album=album,
                license=license,
                global_progress=progressbar,
                remote=ctx.obj["remote"],
                library_or_channel_id=library_or_channel_id,
                container_type="channel" if channel else "library",
                semaphore=sem,
            )
            for path in paths
        ]
        await asyncio.gather(*tasks)
        logs.logger.info("Upload complete")


uploads_rm = base.get_delete_command(uploads, "api/v1/uploads/{}/")
