import json
import uuid

import pytest
import click

from funkwhale_cli import api
from funkwhale_cli import cli


@pytest.fixture
def group():
    @click.group()
    @click.pass_context
    def g():
        pass

    return g


@pytest.fixture
def cli_ctx(mocker):
    c = mocker.Mock(
        obj={
            "remote": api.get_api(
                domain="test.funkwhale", protocol="https", token="test_token"
            )
        },
        color=None,
    )
    click.globals.push_context(c)
    yield c
    click.globals.pop_context()


def test_delete_command(group, cli_ctx, session, responses):
    command = cli.base.get_delete_command(group, "api/v1/noop/{}/")
    id = "fake_id"
    responses.delete("https://test.funkwhale/api/v1/noop/fake_id/")
    command.callback(id=[id], raw=False, no_input=True, _async_reraise=True)
    assert len(responses._responses) == 1


@pytest.mark.parametrize(
    "input, output",
    [
        (
            {"next": "http://test/?page=2", "count": 12, "results": [1, 2]},
            {"current_page": 1, "next_page": 2, "total_pages": 6, "page_size": 2},
        ),
        (
            {"next": "http://test/?page=3", "count": 13, "results": [1, 2]},
            {"current_page": 2, "next_page": 3, "total_pages": 7, "page_size": 2},
        ),
        (
            {"previous": "http://test/?page=6", "count": 13, "results": [1]},
            {"current_page": 7, "next_page": None, "total_pages": 7, "page_size": 2},
        ),
        (
            {"previous": "http://test/", "count": 4, "results": [1, 2, 3, 4]},
            {"current_page": 1, "next_page": None, "total_pages": 1, "page_size": None},
        ),
    ],
)
def test_get_pagination_data(input, output):
    assert cli.base.get_pagination_data(input) == output


def test_lazy_credential(mocker):
    get_password = mocker.patch("keyring.get_password", return_value="password")
    credential = cli.auth.lazy_credential("http://testurl", "_")

    get_password.assert_not_called()

    str(credential)

    get_password.assert_called_once_with("http://testurl", "_")

    assert credential == "password"

    # result is cached
    str(credential)
    assert get_password.call_count == 1


def test_users_me(cli_ctx, session, responses, get_requests):
    command = cli.users.users_me
    url = "https://test.funkwhale/api/v1/users/users/me/"
    responses.get(
        url,
        payload={
            "id": 1,
            "username": "user",
            "full_username": "user@funkwhale.user.com",
            "name": "",
            "email": "contact@user.com",
            "is_staff": True,
            "is_superuser": True,
            "permissions": {"library": True, "moderation": True, "settings": True},
            "date_joined": "2016-04-30T13:36:07.747395Z",
            "privacy_level": "instance",
            "quota_status": {
                "max": 100000,
                "remaining": 82102.249366,
                "current": 17897.750634,
                "skipped": 46.135774,
                "pending": 0.0,
                "finished": 17851.614859999998,
                "errored": 0.0,
            },
        },
    )

    command.callback(raw=False, column=None, format=None)

    requests = get_requests("get", url)
    assert len(requests) == 1


def test_libraries_create(cli_ctx, session, responses, get_requests):
    command = cli.libraries.libraries_create
    url = "https://test.funkwhale/api/v1/libraries/"
    responses.post(url)

    command.callback(name="test", visibility="public", raw=False, _async_reraise=True)

    requests = get_requests("post", url)
    assert len(requests) == 1
    assert requests[0].kwargs["data"] == {"name": "test", "visibility": "public"}


def test_libraries_ls(cli_ctx, session, responses, get_requests):
    command = cli.libraries.libraries_ls
    url = "https://test.funkwhale/api/v1/libraries/?ordering=-creation_date&page=1&page_size=5&q=hello"
    responses.get(
        url, payload={"results": [], "next": None, "previous": None, "count": 0}
    )

    command.callback(
        raw=False,
        page=1,
        page_size=5,
        ordering="-creation_date",
        filter="favorites=true",
        query=["hello"],
        column=None,
        format=None,
        no_headers=False,
        ids=False,
        limit=1,
    )

    requests = get_requests("get", url)
    assert len(requests) == 1


def test_libraries_rm(cli_ctx, session, responses, get_requests):
    command = cli.libraries.libraries_rm
    url = "https://test.funkwhale/api/v1/libraries/"
    responses.delete(url + "1/")
    responses.delete(url + "42/")

    command.callback(id=[1, 42], raw=False, no_input=True, _async_reraise=True)

    assert len(get_requests("delete", url + "1/")) == 1
    assert len(get_requests("delete", url + "42/")) == 1


def test_uploads_create(cli_ctx, session, responses, get_requests, tmpdir):
    tmp_file = tmpdir.join("test.mp3")
    tmp_file.write_text("content", "ascii")
    library_id = str(uuid.uuid4())
    command = cli.uploads.uploads_create

    responses.post("https://test.funkwhale/api/v1/uploads/")
    responses.get("https://test.funkwhale/api/v1/libraries/{}/".format(library_id))

    command.callback(
        library_or_channel_id=library_id,
        paths=[str(tmp_file)],
        parallel=1,
        channel=False,
        ref="test-import",
        draft=False,
        album=None,
        license=None,
        _async_reraise=True,
    )
    expected_data = {
        "library": library_id,
        "import_reference": "test-import",
        "source": "upload://test.mp3",
    }
    upload_requests = get_requests("post", "https://test.funkwhale/api/v1/uploads/")
    assert len(upload_requests) == 1
    data = upload_requests[0].kwargs["data"]
    audio_file = data.pop("audio_file")
    assert data == expected_data
    assert audio_file.name == str(tmp_file)
    libraries_requests = get_requests(
        "get", "https://test.funkwhale/api/v1/libraries/{}/".format(library_id)
    )
    assert len(libraries_requests) == 1


def test_uploads_create_channel(cli_ctx, session, responses, get_requests, tmpdir):
    tmp_file = tmpdir.join("test.mp3")
    tmp_file.write_text("content", "ascii")
    channel_id = str(uuid.uuid4())
    upload_id = str(uuid.uuid4())
    command = cli.uploads.uploads_create
    upload_data = {
        "uuid": upload_id,
        "import_metadata": {
            "title": "test.mp3",
            "album": 12,
            "license": "cc-by-sa-4.0",
        },
    }
    responses.post("https://test.funkwhale/api/v1/uploads/", payload=upload_data)
    responses.get("https://test.funkwhale/api/v1/channels/{}/".format(channel_id))
    responses.get(
        "https://test.funkwhale/api/v1/uploads/{}/audio-file-metadata/".format(
            upload_id
        ),
        payload={"title": "test title"},
    )
    responses.patch("https://test.funkwhale/api/v1/uploads/{}/".format(upload_id))

    command.callback(
        library_or_channel_id=channel_id,
        channel=True,
        paths=[str(tmp_file)],
        parallel=1,
        ref="test-import",
        draft=False,
        album=12,
        license="cc-by-sa-4.0",
        _async_reraise=True,
    )
    expected_data = {
        "channel": channel_id,
        "import_reference": "test-import",
        "source": "upload://test.mp3",
        "import_status": "draft",
        "import_metadata": json.dumps(
            {"title": "test.mp3", "album": 12, "license": "cc-by-sa-4.0"}
        ),
    }
    upload_requests = get_requests("post", "https://test.funkwhale/api/v1/uploads/")
    assert len(upload_requests) == 1
    data = upload_requests[0].kwargs["data"]
    audio_file = data.pop("audio_file")
    assert data == expected_data
    assert audio_file.name == str(tmp_file)
    channels_requests = get_requests(
        "get", "https://test.funkwhale/api/v1/channels/{}/".format(channel_id)
    )
    assert len(channels_requests) == 1
    publish_requests = get_requests(
        "patch", "https://test.funkwhale/api/v1/uploads/{}/".format(upload_id)
    )
    assert len(publish_requests) == 1
    data = publish_requests[0].kwargs["json"]
    expected_data = {
        "import_status": "pending",
        "import_metadata": {
            "title": "test title",
            "album": 12,
            "license": "cc-by-sa-4.0",
        },
    }
    assert data == expected_data


def test_uploads_ls(cli_ctx, session, responses, get_requests):
    command = cli.uploads.uploads_ls
    url = "https://test.funkwhale/api/v1/uploads/?ordering=-creation_date&page=1&page_size=5&q=hello"
    responses.get(
        url, payload={"results": [], "next": None, "previous": None, "count": 0}
    )

    command.callback(
        raw=False,
        page=1,
        page_size=5,
        ordering="-creation_date",
        filter="favorites=true",
        query=["hello"],
        column=None,
        format=None,
        no_headers=False,
        ids=False,
        limit=1,
    )

    requests = get_requests("get", url)
    assert len(requests) == 1


def test_uploads_rm(cli_ctx, session, responses, get_requests):
    uuid1 = str(uuid.uuid4())
    uuid2 = str(uuid.uuid4())
    command = cli.uploads.uploads_rm
    url = "https://test.funkwhale/api/v1/uploads/"
    responses.delete(url + uuid1 + "/")
    responses.delete(url + uuid2 + "/")

    command.callback(id=[uuid1, uuid2], raw=False, no_input=True, _async_reraise=True)

    assert len(get_requests("delete", url + uuid1 + "/")) == 1
    assert len(get_requests("delete", url + uuid2 + "/")) == 1


def test_favorites_tracks_create(cli_ctx, session, responses, get_requests):
    command = cli.favorites.favorites_tracks_create
    url = "https://test.funkwhale/api/v1/favorites/tracks/"
    responses.post(url, repeat=True)

    command.callback(id=[1, 42], _async_reraise=True)

    requests = get_requests("post", url)
    assert len(requests) == 2
    assert requests[0].kwargs["data"] == {"track": 1}
    assert requests[1].kwargs["data"] == {"track": 42}


def test_favorites_tracks_ls(cli_ctx, session, responses, get_requests):
    command = cli.favorites.favorites_tracks_ls
    url = "https://test.funkwhale/api/v1/favorites/tracks/?ordering=-creation_date&page=1&page_size=5&q=hello"
    responses.get(
        url, payload={"results": [], "next": None, "previous": None, "count": 0}
    )

    command.callback(
        raw=False,
        page=1,
        page_size=5,
        ordering="-creation_date",
        filter="favorites=true",
        query=["hello"],
        column=None,
        format=None,
        no_headers=False,
        ids=False,
        limit=1,
    )

    requests = get_requests("get", url)
    assert len(requests) == 1


def test_favorites_tracks_rm(cli_ctx, session, responses, get_requests):
    command = cli.favorites.favorites_tracks_rm
    url = "https://test.funkwhale/api/v1/favorites/tracks/remove/"
    responses.delete(url, repeat=True)

    command.callback(id=[1, 42], _async_reraise=True)

    requests = get_requests("delete", url)
    assert len(requests) == 2
    assert requests[0].kwargs["data"] == {"track": 1}
    assert requests[1].kwargs["data"] == {"track": 42}


def test_tracks_ls(cli_ctx, session, responses, get_requests):
    command = cli.tracks.tracks_ls
    url = "https://test.funkwhale/api/v1/tracks/?ordering=-creation_date&page=1&page_size=5&q=hello"
    responses.get(
        url, payload={"results": [], "next": None, "previous": None, "count": 0}
    )

    command.callback(
        raw=False,
        page=1,
        page_size=5,
        ordering="-creation_date",
        filter="favorites=true",
        query=["hello"],
        column=None,
        format=None,
        no_headers=False,
        ids=False,
        limit=1,
    )

    requests = get_requests("get", url)
    assert len(requests) == 1


def test_tracks_generate_playlist(cli_ctx, session, responses, get_requests, capfd):
    expected = "./Test-Artist - Test-Album - Test-Track.flac\n"

    command = cli.tracks.track_generate_playlist
    url = "https://test.funkwhale/api/v1/tracks/1/"
    responses.get(
        url,
        payload={"uploads": [{"extension": "flac"}], "album": {"title": "Test-Album", "release_date": "2021-02-01"},
                 "artist": {"name": "Test-Artist"}, "title": "Test-Track"}
    )

    command.callback(
        id="1",
        format=None,
        directory=None,
        template="{artist} - {album} - {title}.{extension}",
        base_path="./",
        name="playlist",
    )

    requests = get_requests("get", url)
    assert len(requests) == 1
    out, err = capfd.readouterr()
    assert out == expected

def test_tracks_generate_playlist_base_path(cli_ctx, session, responses, get_requests, capfd):
    expected = "/different_base/Test-Artist - Test-Album - Test-Track.flac\n"

    command = cli.tracks.track_generate_playlist
    url = "https://test.funkwhale/api/v1/tracks/1/"
    responses.get(
        url,
        payload={"uploads": [{"extension": "flac"}], "album": {"title": "Test-Album", "release_date": "2021-02-01"},
                 "artist": {"name": "Test-Artist"}, "title": "Test-Track"}
    )

    command.callback(
        id="1",
        format=None,
        directory=None,
        template="{artist} - {album} - {title}.{extension}",
        base_path="/different_base/",
        name="playlist",
    )

    requests = get_requests("get", url)
    assert len(requests) == 1
    out, err = capfd.readouterr()
    assert out == expected

def test_tracks_generate_playlist_format(cli_ctx, session, responses, get_requests, capfd):
    expected = "./Test-Artist - Test-Album - Test-Track.mp3\n"

    command = cli.tracks.track_generate_playlist
    url = "https://test.funkwhale/api/v1/tracks/1/"
    responses.get(
        url,
        payload={"uploads": [{"extension": "flac"}], "album": {"title": "Test-Album", "release_date": "2021-02-01"},
                 "artist": {"name": "Test-Artist"}, "title": "Test-Track"}
    )

    command.callback(
        id="1",
        format="mp3",
        directory=None,
        template="{artist} - {album} - {title}.{extension}",
        base_path="./",
        name="playlist",
    )

    requests = get_requests("get", url)
    assert len(requests) == 1
    out, err = capfd.readouterr()
    assert out == expected


def test_artists_ls(cli_ctx, session, responses, get_requests):
    command = cli.artists.artists_ls
    url = "https://test.funkwhale/api/v1/artists/?ordering=-creation_date&page=1&page_size=5&q=hello"
    responses.get(
        url, payload={"results": [], "next": None, "previous": None, "count": 0}
    )

    command.callback(
        raw=False,
        page=1,
        page_size=5,
        ordering="-creation_date",
        filter="favorites=true",
        query=["hello"],
        column=None,
        format=None,
        no_headers=False,
        ids=False,
        limit=1,
    )

    requests = get_requests("get", url)
    assert len(requests) == 1


def test_albums_ls(cli_ctx, session, responses, get_requests):
    command = cli.albums.albums_ls
    url = "https://test.funkwhale/api/v1/albums/?ordering=-creation_date&page=1&page_size=5&q=hello"
    responses.get(
        url, payload={"results": [], "next": None, "previous": None, "count": 0}
    )

    command.callback(
        raw=False,
        page=1,
        page_size=5,
        ordering="-creation_date",
        filter="favorites=true",
        query=["hello"],
        column=None,
        format=None,
        no_headers=False,
        ids=False,
        limit=1,
    )

    requests = get_requests("get", url)
    assert len(requests) == 1


def test_playlists_create(cli_ctx, session, responses, get_requests):
    command = cli.playlists.playlists_create
    url = "https://test.funkwhale/api/v1/playlists/"
    responses.post(url)

    command.callback(name="test", visibility="public", raw=False, _async_reraise=True)

    requests = get_requests("post", url)
    assert len(requests) == 1
    assert requests[0].kwargs["data"] == {"name": "test", "privacy_level": "public"}


def test_playlists_ls(cli_ctx, session, responses, get_requests):
    command = cli.playlists.playlists_ls
    url = "https://test.funkwhale/api/v1/playlists/?ordering=-creation_date&page=1&page_size=5&q=hello"
    responses.get(
        url, payload={"results": [], "next": None, "previous": None, "count": 0}
    )

    command.callback(
        raw=False,
        page=1,
        page_size=5,
        ordering="-creation_date",
        filter="favorites=true",
        query=["hello"],
        column=None,
        format=None,
        no_headers=False,
        ids=False,
        limit=1,
    )

    requests = get_requests("get", url)
    assert len(requests) == 1


def test_playlists_ls_mine(
    cli_ctx, session, responses, get_requests, mocker, coroutine_mock
):
    get_user_info = mocker.patch.object(
        cli.base, "get_user_info", coroutine_mock(return_value={"id": 42})
    )
    command = cli.playlists.playlists_ls
    url = "https://test.funkwhale/api/v1/playlists/?ordering=-creation_date&page=1&page_size=5&q=hello&user=42"
    responses.get(
        url, payload={"results": [], "next": None, "previous": None, "count": 0}
    )

    command.callback(
        raw=False,
        page=1,
        page_size=5,
        ordering="-creation_date",
        filter="favorites=true",
        query=["hello"],
        column=None,
        format=None,
        no_headers=False,
        ids=False,
        limit=1,
        owned=True,
    )

    requests = get_requests("get", url)
    assert len(requests) == 1
    get_user_info.assert_called_once_with(cli_ctx)


def test_playlists_rm(cli_ctx, session, responses, get_requests):
    command = cli.playlists.playlists_rm
    url = "https://test.funkwhale/api/v1/playlists/"
    responses.delete(url + "1/")
    responses.delete(url + "42/")

    command.callback(id=[1, 42], raw=False, no_input=True, _async_reraise=True)

    assert len(get_requests("delete", url + "1/")) == 1
    assert len(get_requests("delete", url + "42/")) == 1


def test_playlists_tracks_add(cli_ctx, session, responses, get_requests):
    command = cli.playlists.playlists_tracks_add
    url = "https://test.funkwhale/api/v1/playlists/66/add"
    responses.post(url)

    command.callback(id=66, track=[1, 42], no_duplicates=True, _async_reraise=True)

    requests = get_requests("post", url)
    assert len(requests) == 1
    assert requests[0].kwargs["data"] == {"tracks": [1, 42], "allow_duplicates": False}


def test_playlists_tracks(cli_ctx, session, responses, get_requests):
    command = cli.playlists.playlists_tracks
    url = "https://test.funkwhale/api/v1/playlists/66/tracks/"
    responses.get(url, payload={"results": [], "count": 0})

    command.callback(
        id=66, raw=False, column=None, format=None, no_headers=False, ids=False
    )

    requests = get_requests("get", url)
    assert len(requests) == 1


def test_channel_create_music(cli_ctx, session, responses, get_requests):
    command = cli.channels.channels_create
    url = "https://test.funkwhale/api/v1/channels/"
    responses.post(url)

    command.callback(
        name="test",
        cover="uuid",
        username="hello",
        content_category="music",
        description="description text",
        tags="punk,rock, ska",
        language=None,
        itunes_category=None,
        raw=False,
        _async_reraise=True,
    )

    requests = get_requests("post", url)
    assert len(requests) == 1
    assert requests[0].kwargs["json"] == {
        "name": "test",
        "cover": "uuid",
        "username": "hello",
        "content_category": "music",
        "description": {"text": "description text", "content_type": "text/markdown"},
        "tags": ["punk", "rock", "ska"],
    }


def test_channel_create_podcast(cli_ctx, session, responses, get_requests):
    command = cli.channels.channels_create
    url = "https://test.funkwhale/api/v1/channels/"
    responses.post(url)

    command.callback(
        name="test",
        cover="uuid",
        username="hello",
        content_category="podcast",
        description="description text",
        tags="punk,rock, ska",
        language="en",
        itunes_category="Leisure",
        raw=False,
        _async_reraise=True,
    )

    requests = get_requests("post", url)
    assert len(requests) == 1
    assert requests[0].kwargs["json"] == {
        "name": "test",
        "cover": "uuid",
        "username": "hello",
        "content_category": "podcast",
        "description": {"text": "description text", "content_type": "text/markdown"},
        "tags": ["punk", "rock", "ska"],
        "metadata": {"itunes_category": "Leisure", "language": "en",},
    }


def test_channels_ls(cli_ctx, session, responses, get_requests):
    command = cli.channels.channels_ls
    url = "https://test.funkwhale/api/v1/channels/?ordering=-creation_date&page=1&page_size=5&q=hello&scope=me"
    responses.get(
        url, payload={"results": [], "next": None, "previous": None, "count": 0}
    )

    command.callback(
        raw=False,
        page=1,
        page_size=5,
        ordering="-creation_date",
        filter="subscribed=true",
        scope="me",
        query=["hello"],
        column=None,
        format=None,
        no_headers=False,
        ids=False,
        limit=1,
    )

    requests = get_requests("get", url)
    assert len(requests) == 1


def test_channels_rm(cli_ctx, session, responses, get_requests):
    command = cli.channels.channels_rm
    url = "https://test.funkwhale/api/v1/channels/"
    responses.delete(url + "uuid1/")
    responses.delete(url + "uuid2/")

    command.callback(
        id=["uuid1", "uuid2"], raw=False, no_input=True, _async_reraise=True
    )

    assert len(get_requests("delete", url + "uuid1/")) == 1
    assert len(get_requests("delete", url + "uuid2/")) == 1
